#include "Burden.h"
#include "Constant.h"
#include "math.h"
#include "MathStats.h"
#include <string>
#include <cstring>
#include <iostream>
#include <stdlib.h>
#include "qfc.h"
#include "MixChidist.h"
#define MATHLIB_STANDALONE 
#include <Rmath.h>
#include <Eigen/SVD>

String Burden::genename = "";
String Burden::genename_f = "";
String Burden::outputFile = "";
double Burden::mafMin = 0.0;
double Burden::mafMax = 0.5;
double Burden::MAC = 0.0;
String Burden::chromosome = "";
bool Burden::vcfAnnotated = true;
bool Burden::CMC_binary = false;
bool Burden::CMC_counts = false;
bool Burden::MB = false;
bool Burden::burden = false;
bool Burden::LinScoreAssoc = false;
bool Burden::SCORE=true;
bool Burden::LRT = false;
bool Burden::WALD = false;
bool Burden::SKAT = false;
bool Burden::fullResult = true;

Burden::Burden(){}

Burden::~Burden()
{}

bool Burden::GetGenotypeVectorDosage(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum) 
{
   bool status = false;
   varsIncluded = "";

   //mac is the total number of alternative alleles included in the group
   mac = 0.0;
   //numVarPerGene is the total number of variants included in this gene.
   numVarPerGene=0;

   mafIncluded.Clear();
   Vector mafSqrIncluded;

   reader.set1BasedReadSection(group.chrList[groupNum][0].c_str(),group.posList[groupNum][0],group.posList[groupNum][group.posList[groupNum].Length()-1]+1);
   int m=0;
   while(reader.readRecord(record))
   {
      int position = record.get1BasedPosition();
      String altAllele = record.getAltStr();
      String refAllele = record.getRefStr();
      if(group.posList[groupNum][m]!=position)
      {
	 continue;
      }

      String allele = refAllele + "/" + altAllele;
      if(allele != group.alleleList[groupNum][m])
      {
	 String flipAllele = altAllele + "/" + refAllele;
	 if(flipAllele != group.alleleList[groupNum][m])
	 {
	    continue;
	 }
	 printf("ref and alt alleles from group files are flipped from vcf file!\n");
      }

      //If the variant match the record in groupfile then add up m to next variant. 
      //This requires varaints in group file has to be sorted.
      m++;
      if(m==group.chrList[groupNum].Length()+1)
	 break;

      int n=0,nf=0;
      double fmaf = 0.0, maf=0.0;
      VcfRecordGenotype & genoInfo = record.getGenotypeInfo();

      bool XStatus = false;
      if(AutoFit::fitX)
      {
	 String chr = group.chrList[groupNum][0].c_str();
	 if(chr==SVT::xLabel && position > SVT::Xstart && position < SVT::Xend)
	 {
	    XStatus=true;
	 }
      }

      Vector genotype;   
      genotype.Dimension(trans.persons);
      for(int i=0;i<trans.persons;i++)
	 genotype[i]=-1.0;

      //Loop through samples according to trans.pPheno 
      //read in dosage into genotype vector
      int idx=0;
      for (int f = 0; f < ped.familyCount; f++)
      {
	 for(int j=0;j<trans.pPheno[f].Length();j++)
	 {
	    String sample = ped[trans.pPheno[f][j]].pid;

	    //get the sample number by sample ID from hash
	    int s = trans.sampleVCFIDHash.Integer(sample);
	    if(s!=-1)
	    {
	       const std::string * geno = genoInfo.getString("DS",s);
	       //if dosage is not missing
	       if(strcmp((*geno).c_str(),".")>0)
	       {
		  float dose = atof((*geno).c_str());
		  if(dose >=0)
		  {
		     if(XStatus && dose >0 && ped[trans.pPheno[f][j]].sex==SVT::maleLabel)
			genotype[idx] = 2;
		     else
			genotype[idx] = dose;

		     maf += genotype[idx];
		     n++;

		     //judge if this is a founder
		     int founder = trans.foundersHash.Integer(sample);
		     if(founder != -1)
		     {
			fmaf += genotype[idx];
			nf++;
		     }
		  }
	       }
	    }
	    else
	    {
	       error("FATAL ERROR! Sample %s is not included in VCF file.\n",sample.c_str());
	    }
	    idx++;
	 }
      }
      maf /= 2.0*n;
      fmaf /= 2.0*nf;

      if(fmaf==0.0 && maf>0)
	 fmaf = 1.0/(2.0*nf);

      //check if frequencies are read from outside
      if(!group.freqOut)
      {
	 //check if maf falls into the range
	 if(SVT::FounderFreq)
	 {
	    double fmaf_tmp;
	    fmaf_tmp = fmaf;
	    if(fmaf>0.5)
	       fmaf_tmp = 1.0-fmaf;
	    if(fmaf_tmp>mafMax || fmaf_tmp<mafMin)
	       continue;
	 }
	 else
	 {
	    double maf_tmp;
	    maf_tmp=maf;
	    if(maf>0.5)
	       maf_tmp = 1.0-maf;
	    if(maf_tmp>mafMax || maf_tmp<mafMin)
	       continue;
	 }
      }

      if(maf>0. && maf!=1.0)
      {
	 status=true;
      }
      else
      {
	 continue;
      }
      numVarPerGene++;

      bool flip=false;
      if(SVT::FounderFreq && fmaf>0.5)
      {
	 flip=true;
      }
      else if(maf>0.5)
      {
	 flip=true;
      }
      if(flip)
      {
	 fmaf = 1.0-fmaf;
	 maf = 1.0-maf;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {
	       genotype[i] = 2.0-genotype[i];
	    }
	 }
      }

      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {
	    mac+=genotype[i];
	 }
      }

      if(fullResult)
      {
	 String tmp = "";
	 tmp += group.chrList[groupNum][m-1];
	 tmp+=":";
	 tmp+=group.posList[groupNum][m-1];
	 tmp+= "_";
	 if(flip)
	 {
	    tmp+=altAllele;
	    tmp+="/";
	    tmp += refAllele;
	 }
	 else
	 {
	    tmp+=refAllele;
	    tmp+="/";
	    tmp += altAllele;
	 }
	 tmp+=  ";";
	 varsIncluded += tmp;
      }

      if(SVT::FounderFreq)
      {
	 mafIncluded.Push(fmaf);
	 mafSqrIncluded.Push(fmaf*fmaf);
      }
      else 
      {
	 mafIncluded.Push(maf);
	 mafSqrIncluded.Push(maf*maf);
      }

      //update genotype vector and genotype matrix
      if(SKAT || fullResult)
      {
	 genotypeSKAT.GrowTo(trans.persons,numVarPerGene);
	 for(int row=0;row<trans.persons;row++)
	 {
	    genotypeSKAT[row][numVarPerGene-1] = genotype[row];
	 }
      }

      if(CMC_counts)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]==-1.0)
	       genotypeCMC_c[row] += 2.0*maf;
	    else
	       genotypeCMC_c[row] += genotype[row];
	 }
      }

      if(CMC_binary)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]>0 && genotypeCMC_b[row]==0)
	    {
	       genotypeCMC_b[row] = 1.0;
	    }
	 }
      }

      if(MB)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]==-1.0)
	    {
	       if(SVT::FounderFreq)
	       {
		  genotypeMB[row] += 2.0*fmaf/sqrt(fmaf*(1.0-fmaf));
	       }
	       else
	       {
		  genotypeMB[row] += 2.0*maf/sqrt(maf*(1.0-maf));
	       }
	    }
	    else
	    {
	       if(SVT::FounderFreq)
	       {     
		  genotypeMB[row] += genotype[row]/sqrt(fmaf*(1.0-fmaf));
	       }     
	       else  
	       {     
		  genotypeMB[row] += genotype[row]/sqrt(maf*(1.0-maf));
	       }
	    }
	 }
      }
   }

   if(CMC_counts)
   {
      double mean = genotypeCMC_c.Average();
      genotypeCMC_c.Add(-mean);
   }
   if(MB)
   {
      double mean = genotypeMB.Average();
      genotypeMB.Add(-mean);
   }

   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }
   freq_min=mafIncluded.Min();
   freq_max=mafIncluded.Max();
   freq_mean=mafIncluded.Sum()/numVarPerGene;
   freq_se = mafSqrIncluded.Sum()/numVarPerGene - freq_mean*freq_mean; 
   //printf("There are %d variants included in the gene\n",numVarPerGene);

   if(mac<MAC) 
      status = false;
   return status;
}

bool Burden::GetGenotypeVectorVCF(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;

   varsIncluded = "";
   //mac is the total number of alternative alleles included
   mac = 0.0;
   //numVarPerGene is the total number of variants included in this gene.
   numVarPerGene=0;

   mafIncluded.Clear();
   Vector mafSqrIncluded;

   reader.set1BasedReadSection(group.chrList[groupNum][0].c_str(),group.posList[groupNum][0],group.posList[groupNum][group.posList[groupNum].Length()-1]+1);
   int m=0;
   while(reader.readRecord(record))
   {
      int position = record.get1BasedPosition();
      String altAllele = record.getAltStr();
      String refAllele = record.getRefStr();

      bool XStatus = false;

      if(AutoFit::fitX)
      {
	 String chr = group.chrList[groupNum][0].c_str();
	 if(chr==SVT::xLabel && position > SVT::Xstart && position < SVT::Xend)
	 {
	    XStatus=true;
	 }
      }

      if(group.posList[groupNum][m]!=position)
      {
	 continue;
      }

      String allele = refAllele + "/" + altAllele;
      if(allele != group.alleleList[groupNum][m])
      {  
	 String flipAllele = altAllele + "/" + refAllele;
	 if(flipAllele != group.alleleList[groupNum][m])
	 {  
	    continue;
	 }  
	 printf("ref and alt alleles from group files are flipped from vcf file!\n");
      }

      //If the variant match the record in groupfile then add up m to next variant. 
      //This requires varaints in group file has to be sorted.
      m++;
      if(m==group.chrList[groupNum].Length()+1)
	 break;

      double fmaf=0.0,maf=0.0;
      int n=0,nf=0;

      //Fill in genotype vector
      Vector genotype;
      genotype.Dimension(trans.persons);
      for(int i=0;i<trans.persons;i++)
	 genotype[i] = -1.0;

      int idx=0;
      for (int f = 0; f < ped.familyCount; f++)
      {
	 for(int p=0;p<trans.pPheno[f].Length();p++)
	 {
	    String sample = ped[trans.pPheno[f][p]].pid.c_str();

	    //get the sample number by sample ID from hash
	    int s = trans.sampleVCFIDHash.Integer(sample);
	    if(s!=-1)
	    {
	       int numGTs = record.getNumGTs(s);
	       Vector allele;

	       for(int j = 0; j < numGTs; j++)
	       {
		  allele.Push(record.getGT(s,j));
	       }
	       int bad =0;
	       for(int m=0;m<allele.Length();m++)
	       {
		  bad += (allele[m]==VcfGenotypeSample::MISSING_GT);
	       }
	       //if no missing allele exists
	       //if no missing allele exists
	       if(bad==0)
	       {
		  int sum=0;
		  for(int i=0;i<allele.Length();i++)
		     sum+=allele[i];

		  if(XStatus && sum >0 && ped[trans.pPheno[f][p]].sex==SVT::maleLabel)
		     genotype[idx] = 2;
		  else
		     genotype[idx] = sum;

		  maf += genotype[idx];
		  n++;

		  //judge if this is a founder
		  int founder = trans.foundersHash.Integer(sample);
		  if(founder!=-1)
		  {
		     nf++;
		     fmaf+=genotype[idx];
		  }
	       }
	    }
	    idx++;
	 }
      }

      //Adjust genotype here
      maf /= 2.0*n;
      fmaf /=2.0*nf;

      if(fmaf==0.0 && maf>0)
	 fmaf = 1.0/(2.0*nf);

      //check if maf falls into the range
      if(!group.freqOut)
      {
	 if(SVT::FounderFreq)
	 {
	    double fmaf_tmp;
	    fmaf_tmp = fmaf;
	    if(fmaf>0.5)
	       fmaf_tmp = 1.0-fmaf;
	    if(fmaf_tmp>mafMax || fmaf_tmp<mafMin)
	       continue;
	 }
	 else
	 {
	    double maf_tmp;
	    maf_tmp=maf;
	    if(maf>0.5)
	       maf_tmp = 1.0-maf;
	    if(maf_tmp>mafMax || maf_tmp<mafMin)
	       continue;
	 }
      }

      if(maf>0. && maf!=1.0)
      {
	 status=true;
      }
      else
      {
	 continue;
      }
      numVarPerGene++;

      bool flip=false;
      if(SVT::FounderFreq && fmaf>0.5)
      {
	 flip=true;
      }
      else if(maf>0.5)
      {
	 flip=true;
      }
      if(flip)
      {
	 fmaf = 1.0-fmaf;
	 maf = 1.0-maf;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {
	       genotype[i] = 2.0-genotype[i];
	    }
	 }
      }

      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {  
	    mac+=genotype[i];
	 }  
      }

      if(fullResult)
      {
	 String tmp;
	 tmp += group.chrList[groupNum][m-1];
	 tmp+=":";
	 tmp+=group.posList[groupNum][m-1];
	 tmp+= "_";
	 if(flip)
	 {
	    tmp+=altAllele;
	    tmp+="/";
	    tmp += refAllele;
	 }
	 else
	 {
	    tmp+=refAllele;
	    tmp+="/";
	    tmp += altAllele;
	 }
	 tmp+=  ";";
	 varsIncluded += tmp;
      }

      if(SVT::FounderFreq)
      {
	 mafIncluded.Push(fmaf);
	 mafSqrIncluded.Push(fmaf*fmaf);
      }
      else
      {
	 mafIncluded.Push(maf);
	 mafSqrIncluded.Push(maf*maf);
      }

      //update genotype vector and genotype matrix
      if(SKAT || fullResult)
      {
	 genotypeSKAT.GrowTo(trans.persons,numVarPerGene);
	 for(int row=0;row<trans.persons;row++)
	 {
	    genotypeSKAT[row][numVarPerGene-1] = genotype[row];
	 }
      }

      if(CMC_counts)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]==-1.0)
	       genotypeCMC_c[row] += 2.0*maf;
	    else
	       genotypeCMC_c[row] += genotype[row];
	 }
      }

      if(CMC_binary)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]>0 && genotypeCMC_b[row]==0)
	    {
	       genotypeCMC_b[row] = 1.0;
	    }
	 }
      }

      if(MB)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]==-1.0)
	    {
	       if(SVT::FounderFreq)
	       {
		  genotypeMB[row] += 2.0*fmaf/sqrt(fmaf*(1.0-fmaf));
	       }
	       else
	       {
		  genotypeMB[row] += 2.0*maf/sqrt(maf*(1.0-maf));
	       }
	    }
	    else
	    {
	       if(SVT::FounderFreq)
	       {
		  genotypeMB[row] += genotype[row]/sqrt(fmaf*(1.0-fmaf));
	       }
	       else
	       {
		  genotypeMB[row] += genotype[row]/sqrt(maf*(1.0-maf));
	       }
	    }
	 }
      }
   }

   if(CMC_counts)
   {
      double mean = genotypeCMC_c.Average();
      genotypeCMC_c.Add(-mean);
   }
   if(MB)
   {
      double mean = genotypeMB.Average();
      genotypeMB.Add(-mean);
   }

   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }

   freq_min=mafIncluded.Min();
   freq_max=mafIncluded.Max();
   freq_mean=mafIncluded.Sum()/numVarPerGene;
   freq_se = mafSqrIncluded.Sum()/numVarPerGene - freq_mean*freq_mean;

   if(mac<MAC)
      status = false;
   return status;
}

bool Burden::GetGenotypeVectorPED(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;
   varsIncluded = "";

   //mac is the total number of alternative alleles included in the group
   mac = 0.0;
   //numVarPerGene is the total number of variants included in this gene.
   numVarPerGene=0;

   mafIncluded.Clear();
   Vector mafSqrIncluded;

   for(int m=0;m<group.chrList[groupNum].Length();m++)
   {
      String markerName = group.chrList[groupNum][m];
      markerName += ":";
      markerName += group.posList[groupNum][m];
      int markerid = ped.markerLookup.Integer(markerName);
      if(markerid==-1)
      {
	 continue;
      }
      int alleleCount = CountAlleles(ped, markerid);
      if(alleleCount!=2)
	 continue;

      String refAllele = ped.GetMarkerInfo(markerid)->alleleLabels[1];
      String altAllele = ped.GetMarkerInfo(markerid)->alleleLabels[2];

      String allele = refAllele + "/" + altAllele;
      if(allele != group.alleleList[groupNum][m])
      {
	 String flipAllele = altAllele + "/" + refAllele;
	 if(flipAllele != group.alleleList[groupNum][m])
	 {
	    continue;
	 }
      }

      //if marker m from group list is found then get the genotype vector
      bool XStatus = false;
      if(AutoFit::fitX)
      {
	 int found = markerName.FastFind(SVT::xLabel,0);
	 if(found!=-1)
	 {
	    if(group.posList[groupNum][m] > SVT::Xstart && group.posList[groupNum][m] < SVT::Xend)
	    {
	       XStatus=true;
	    }
	 }
      }

      int n=0,nf=0;
      double fmaf = 0.0, maf=0.0;
      Vector genotype;
      genotype.Dimension(trans.persons);
      for(int i=0;i<trans.persons;i++)
	 genotype[i]=-1.0;

      //Loop through samples according to trans.pPheno 
      int idx=0;
      for(int f=0;f<ped.familyCount;f++)
      {
	 for(int p=0;p<trans.pPheno[f].Length();p++)
	 {
	    if(!ped[trans.pPheno[f][p]].isGenotyped(markerid))
	    {
	       idx++;
	       continue;
	    }
	    if(ped[trans.pPheno[f][p]].markers[markerid].isHeterozygous())
	    {
	       if(XStatus)
	       {
		  if(ped[trans.pPheno[f][p]].sex==SVT::femaleLabel)
		  {
		     genotype[idx] = 1.0;
		  }
		  if(ped[trans.pPheno[f][p]].sex==SVT::maleLabel)
		  {
		     genotype[idx] = 2.0;
		  }
	       }
	       else
	       {
		  genotype[idx] = 1.0;
	       }
	    }
	    else if(ped[trans.pPheno[f][p]].markers[markerid].isHomozygous())
	    {
	       if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(2))
	       {
		  genotype[idx] = 2.0;
	       }
	       else if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(1))
	       {
		  genotype[idx] = 0.0;
	       }
	    }
	    maf+=genotype[idx];
	    n++;
	    int founder = trans.foundersHash.Integer(ped[trans.pPheno[f][p]].pid);
	    if(founder != -1)
	    {
	       fmaf += genotype[idx];
	       nf++;
	    }
	    idx++;
	 }
      }

      maf /= 2.0*n;
      fmaf /= 2.0*nf;

      if(fmaf==0.0 && maf>0)
	 fmaf = 1.0/(2.0*nf);

      //check if maf falls into the range
      if(!group.freqOut)
      {
	 if(SVT::FounderFreq)
	 {
	    double fmaf_tmp;
	    fmaf_tmp = fmaf;
	    if(fmaf>0.5)
	       fmaf_tmp = 1.0-fmaf;
	    if(fmaf_tmp>mafMax || fmaf_tmp<mafMin)
	       continue;
	 }
	 else
	 {
	    double maf_tmp;
	    maf_tmp=maf;
	    if(maf>0.5)
	       maf_tmp = 1.0-maf;
	    if(maf_tmp>mafMax || maf_tmp<mafMin)
	       continue;
	 }
      }

      if(maf>0. && maf!=1.0)
      {
	 status=true;
      }
      else
      {
	 continue;
      }
      numVarPerGene++;

      bool flip=false;
      if(SVT::FounderFreq && fmaf>0.5)
      {
	 flip=true;
      }
      else if(maf>0.5)
      {
	 flip=true;
      }
      if(flip)
      {
	 fmaf = 1.0-fmaf;
	 maf = 1.0-maf;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {
	       genotype[i] = 2.0-genotype[i];
	    }
	 }
      }

      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {
	    mac+=genotype[i];
	 }
      }

      if(fullResult)
      {
	 String tmp = "";
	 tmp += group.chrList[groupNum][m];
	 tmp+=":";
	 tmp+=group.posList[groupNum][m];
	 tmp+= "_";
	 if(flip)
	 {
	    tmp+=altAllele;
	    tmp+="/";
	    tmp += refAllele;
	 }
	 else
	 {
	    tmp+=refAllele;
	    tmp+="/";
	    tmp += altAllele;
	 }
	 tmp+=  ";";
	 varsIncluded += tmp;
      }

      if(SVT::FounderFreq)
      {
	 mafIncluded.Push(fmaf);
	 mafSqrIncluded.Push(fmaf*fmaf);
      }
      else
      {
	 mafIncluded.Push(maf);
	 mafSqrIncluded.Push(maf*maf);
      }

      //update genotype vector and genotype matrix
      if(SKAT || fullResult)
      {
	 genotypeSKAT.GrowTo(trans.persons,numVarPerGene);
	 for(int row=0;row<trans.persons;row++)
	 {
	    genotypeSKAT[row][numVarPerGene-1] = genotype[row];
	 }
      }

      if(CMC_counts)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]==-1.0)
	       genotypeCMC_c[row] += 2.0*maf;
	    else
	       genotypeCMC_c[row] += genotype[row];
	 }
      }

      if(CMC_binary)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]>0 && genotypeCMC_b[row]==0)
	    {
	       genotypeCMC_b[row] = 1.0;
	    }
	 }
      }

      if(MB)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]==-1.0)
	    {
	       if(SVT::FounderFreq)
	       {
		  genotypeMB[row] += 2.0*fmaf/sqrt(fmaf*(1.0-fmaf));
	       }
	       else
	       {
		  genotypeMB[row] += 2.0*maf/sqrt(maf*(1.0-maf));
	       }
	    }
	    else
	    {
	       if(SVT::FounderFreq)
	       {
		  genotypeMB[row] += genotype[row]/sqrt(fmaf*(1.0-fmaf));
	       }
	       else
	       {
		  genotypeMB[row] += genotype[row]/sqrt(maf*(1.0-maf));
	       }
	    }
	 }
      }
   }
   if(CMC_counts)
   {
      double mean = genotypeCMC_c.Average();
      genotypeCMC_c.Add(-mean);
   }
   if(MB)
   {
      double mean = genotypeMB.Average();
      genotypeMB.Add(-mean);
   }

   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }
   freq_min=mafIncluded.Min();
   freq_max=mafIncluded.Max();
   freq_mean=mafIncluded.Sum()/numVarPerGene;
   freq_se = mafSqrIncluded.Sum()/numVarPerGene - freq_mean*freq_mean;
   //printf("There are %d variants included in the gene\n",numVarPerGene);

   if(mac<MAC)
      status = false;
   return status;
}

bool Burden::GetGenotypeVector(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status;

   if(group.chrList[groupNum].Length()==0)
   {
      status = false;
      return status;
   }

   if(MB)
   {
      genotypeMB.Dimension(trans.persons);
      genotypeMB.Zero();
   }
   if(CMC_binary)
   {
      genotypeCMC_b.Dimension(trans.persons);
      genotypeCMC_b.Zero();
   }
   if(CMC_counts)
   {
      genotypeCMC_c.Dimension(trans.persons);
      genotypeCMC_c.Zero();
   }
   if(SKAT || fullResult)
      genotypeSKAT.Dimension(trans.persons,1,-1.0);

   if(SVT::genoFromPed)
      status = GetGenotypeVectorPED(trans,ped,group,groupNum);
   if(SVT::genoFromVCF)
      status = GetGenotypeVectorVCF(trans,ped,group,groupNum);
   if(SVT::dosage)
      status = GetGenotypeVectorDosage(trans,ped,group,groupNum);

   if(!status)
      return status;

   if(SKAT || fullResult)
   {
      //adjust genotype matrix by mean
      for(int col=0;col<genotypeSKAT.cols;col++)
      {
	 double mean = 0.0;
	 int n_=0;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotypeSKAT[i][col]!=-1.0)
	    {
	       mean += genotypeSKAT[i][col];
	       n_++;
	    }
	 }
	 mean /= n_;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotypeSKAT[i][col]==-1.0)
	       genotypeSKAT[i][col]=0.0;
	    else
	       genotypeSKAT[i][col] -= mean;
	 }
      }
   }
   //printf("start refinning SKAT genotype matrix.\n");
   //refine the genotype matrix for SKAT
   //delete duplicate markers 

   del_status.Clear();

   if(!SKAT)
      return status;

   if(SKAT && genotypeSKAT.cols >1)
   {
      Vector cp_mafIncluded,mask;
      cp_mafIncluded.Copy(mafIncluded);
      cp_mafIncluded.Sort();
      mask.Dimension(mafIncluded.Length(),0);
      for(int i=1;i<mask.Length();i++)
      {
	 if(cp_mafIncluded[i]==cp_mafIncluded[i-1])
	    mask[i] = 1;
      }
      //printf("mask has sum of %g\n",mask.Sum());

      if(mask.Sum()>0)
      {
	 del_status.Dimension(genotypeSKAT.cols,0.0);
	 //find duplicate sums
	 Vector dup,tmp;
	 tmp.Dimension(genotypeSKAT.cols,0);
	 for(int s=0;s<mask.Length();s++)
	 {
	    if(mask[s]==1) 
	    {
	       dup.Push(cp_mafIncluded[s]);
	    }
	 }
	 dup.RemoveDuplicates();

	 for(int d=0;d<dup.Length();d++)
	 {
	    Vector colnum;
	    for(int i=0;i<mafIncluded.Length();i++)
	    {
	       if(mafIncluded[i]==dup[d])
		  colnum.Push(i);
	    }
	    if(colnum.Length()==0)
	    {
	       printf("Error! Should have duplicate sum values. colnum should have lenght>0.\n");
	       continue;
	    }

	    //now look at those columns of genotypes to check if they are perfect             //replicates to each other
	    for(int i=0;i<colnum.Length();i++)
	    {
	       int start_col;
	       int tmp = colnum[i];
	       if(del_status[tmp]==0)
		  start_col = colnum[i];
	       else 
		  continue;
	       for(int col=i+1;col<colnum.Length();col++)
	       {
		  int col_idx = colnum[col];
		  del_status[col_idx] = 1;
		  for(int row=0;row<genotypeSKAT.rows;row++)
		  {
		     if(genotypeSKAT[row][start_col]!=genotypeSKAT[row][col_idx])
		     {
			del_status[col_idx]=0;
			break;
		     }
		  }
	       }
	    }
	 }
	 //adjust genotype matrix by mean
	 /*
	    for(int col=0;col<genotypeSKAT.cols;col++)
	    {
	    double mean = 0.0;
	    int n_=0;
	    for(int i=0;i<genotypeSKAT.rows;i++)
	    {
	    if(genotypeSKAT[i][col]!=-1.0)
	    {
	    mean += genotypeSKAT[i][col];
	    n_++;
	    }
	    }
	    mean /= n_;
	    for(int i=0;i<genotypeSKAT.rows;i++)
	    {
	    if(genotypeSKAT[i][col]==-1.0)
	    genotypeSKAT[i][col]=0.0;
	    else
	    genotypeSKAT[i][col] -= mean;
	    }
	    }
	  */
      }
   }
   return status;
}

void Burden::Calculate(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group,FILE * log)
{
   int traitNum = engine.GetTraitID(ped,traitName.c_str(),log);
   FILE * MBScoreOut;
   FILE * CMC_bScoreOut;
   FILE * CMC_cScoreOut;
   FILE * MBLRTOut;
   FILE * CMC_bLRTOut;
   FILE * CMC_cLRTOut;
   FILE * MBWALDOut;
   FILE * CMC_bWALDOut;
   FILE * CMC_cWALDOut;
   FILE * MBLinScoreOut;
   FILE * CMC_bLinScoreOut;
   FILE * CMC_cLinScoreOut;
   FILE * SKATOut;

   if(SCORE)
   {
      if(MB) 
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.score.txt";
	 MBScoreOut = fopen(MBfilename,"wt");
	 PrintHeader(MBScoreOut ,trans,engine,traitName);
      }

      if(CMC_binary) 
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.score.txt";
	 CMC_bScoreOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bScoreOut ,trans,engine,traitName);
      }

      if(CMC_counts) 
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".CMC_counts.score.txt";
	 CMC_cScoreOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cScoreOut,trans,engine,traitName);
      }
   }

   if(WALD)
   {
      if(MB)
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.wald.txt";
	 MBWALDOut = fopen(MBfilename,"wt");
	 PrintHeader(MBWALDOut ,trans,engine,traitName);
      }

      if(CMC_binary)
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.wald.txt";
	 CMC_bWALDOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bWALDOut ,trans,engine,traitName);
      }

      if(CMC_counts)
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".CMC_counts.wald.txt";
	 CMC_cWALDOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cWALDOut,trans,engine,traitName);
      }
   }

   if(LinScoreAssoc)
   {
      if(MB)
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.LinScore.txt";
	 MBLinScoreOut = fopen(MBfilename,"wt");
	 PrintHeader(MBLinScoreOut ,trans,engine,traitName);
      }

      if(CMC_binary)
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.LinScore.txt";
	 CMC_bLinScoreOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bLinScoreOut ,trans,engine,traitName);
      }

      if(CMC_counts)
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".CMC_counts.LinScore.txt";
	 CMC_cLinScoreOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cLinScoreOut,trans,engine,traitName);
      }
   }

   if(LRT)
   {
      if(MB)
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.LRT.txt";
	 MBLRTOut = fopen(MBfilename,"wt");
	 PrintHeader(MBLRTOut,trans,engine,traitName);
      }

      if(CMC_binary)
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.LRT.txt";
	 CMC_bLRTOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bLRTOut,trans,engine,traitName);
      }

      if(CMC_counts)
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".CMC_counts.LRT.txt";
	 CMC_cLRTOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cLRTOut,trans,engine,traitName);
      }
   }

   if(SKAT)
   {
      String SKATfilename;
      SKATfilename = outputFile + traitName + ".SKAT.txt";
      SKATOut = fopen(SKATfilename,"wt");
      PrintSKATHeader(SKATOut,trans,engine,traitName);
   }

   double L0,L1;

   if(LRT)
   {
      if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
      {
	 L0=engine.EvaluateMulti(trans,false);
      }
      else
      {
	 L0 = engine.Evaluate(engine.deltaHat,trans,false);
	 //printf("L0 is:%g\n",L0);
      }
   }

   int bad =0;
   if(SVT::dosage || SVT::genoFromVCF)
   {
      reader.open(GroupFromAnnotation::vcfInput,header);
      reader.readVcfIndex();
   }
   for(int i=0;i<group.annoGroups.Length();i++)
   {
      if(i%1000==1 && i>1000)
      {
	 printf("Finished analyzing %d genes.\n",((int)i/1000)*1000);
      }
      String geneName = group.annoGroups[i];
      //printf("now working on gene %s\n",geneName.c_str());
      if(GetGenotypeVector(trans,ped,group,i))
      {
	 //printf("working on gene %s\n",geneName.c_str());
	 if(fullResult)
	 {
	    CalculateSingleVarPvalue(trans,geneName,engine,genotypeSKAT,ped);
	 }
	 Vector transGenoMB;
	 Vector transGenoCMC_b;
	 Vector transGenoCMC_c;

	 if(MB)
	    //   transGenoMB.Product(trans.transU,genotypeMB);
	    transGenoMB.Product(trans.UD,genotypeMB);
	 if(CMC_binary)
	    //transGenoCMC_b.Product(trans.transU,genotypeCMC_b);
	    transGenoCMC_b.Product(trans.UD,genotypeCMC_b);
	 if(CMC_counts)
	    //transGenoCMC_c.Product(trans.transU,genotypeCMC_c);
	    transGenoCMC_c.Product(trans.UD,genotypeCMC_c);

	 if(LRT || WALD)
	 {
	    FastTransform  fullTrans;
	    FastFit fullEngine;
	    fullTrans.persons = trans.persons;
	    fullTrans.fixedEffects = trans.fixedEffects +1;

	    if(AutoFit::fitX)
	    {
	       int covariateId = ped.GetCovariateID("MarkerTested");

	       fullTrans.pPheno = new IntArray [ped.familyCount];
	       for (int f = 0; f < ped.familyCount; f++)
	       {
		  fullTrans.pPheno[f].Dimension(0);
		  for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
		  {
		     if(ped[i].isPhenotyped(traitNum) && (!engine.useCovariates || ped[i].isFullyControlled()))
		     {
			fullTrans.pPheno[f].Push(i);
		     }
		  }
	       }
	       int index =0;
	       for (int f = 0; f < ped.familyCount; f++)
	       {
		  if (fullTrans.pPheno[f].Length() > 0)
		  {
		     int count = fullTrans.pPheno[f].Length();
		     for(int i=0;i<count;i++)
		     {
			if(MB)
			   ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotypeMB[index];
			if(CMC_binary)
			   ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotypeCMC_b[index];
			if(CMC_counts)
			   ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotypeCMC_c[index];
			index++;
		     }
		  }
	       }
	       printf("ped.covariateCount is%d\n",ped.covariateCount);

	       AutoFit fullAutoEngine;
	       fullAutoEngine.useCovariates = true;
	       fullAutoEngine.fitSharedEnvironment = AutoFit::fitSharedEnvironment;
	       fullAutoEngine.fitX = AutoFit::fitX;

	       //Fit the alternative
	       fullAutoEngine.FitPolygenicModels(ped,traitNum,true);

	       fullTrans.Y.Copy(trans.Y);
	       fullTrans.X.Copy(trans.X);
	       fullTrans.X.GrowTo(fullTrans.persons,trans.X.cols+1,0.0);
	       fullTrans.UY.Dimension(fullTrans.persons);
	       fullTrans.UX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
	       fullEngine.beta.Dimension(fullTrans.fixedEffects);
	       fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
	       fullTrans.transU.Dimension(0,0);
	       fullTrans.D.Dimension(0);

	       if(MB)
		  for(int i=0;i<fullTrans.persons;i++)
		     fullTrans.X[i][trans.fixedEffects] = genotypeMB[i];
	       if(CMC_binary)
		  for(int i=0;i<fullTrans.persons;i++)
		     fullTrans.X[i][trans.fixedEffects] = genotypeCMC_b[i];
	       if(CMC_counts)
		  for(int i=0;i<fullTrans.persons;i++)
		     fullTrans.X[i][trans.fixedEffects] = genotypeCMC_c[i];

	       fullTrans.traitVar = fullAutoEngine.variance;
	       fullTrans.TransformPedkinX(fullAutoEngine, ped);
	       fullEngine.betaHat.Copy(fullAutoEngine.betaHat);
	    }
	    else 
	    {
	       fullTrans.D.Copy(trans.D);
	       fullTrans.UY.Copy(trans.UY);
	       fullTrans.transU.Copy(trans.transU);
	       fullTrans.UD.Dimension(fullTrans.persons,fullTrans.persons);

	       fullEngine.beta.Dimension(fullTrans.fixedEffects);
	       fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
	       fullTrans.UX.Copy(trans.UX);
	       fullTrans.UX.GrowTo(trans.UX.rows,trans.UX.cols+1,0.0);
	    }

	    String method;
	    if(LRT)
	    {
	       if(MB)
	       {
		  method = "MB";
		  MerlinLRT(L0,fullTrans,MBLRTOut,transGenoMB,geneName,fullEngine,method);
	       }
	       if(CMC_binary)
	       {
		  method = "CMC_binary";
		  MerlinLRT(L0,fullTrans,CMC_bLRTOut,transGenoCMC_b,geneName,fullEngine,method);
	       }
	       if(CMC_counts)
	       {
		  method = "CMC_counts";
		  MerlinLRT(L0,fullTrans,CMC_cLRTOut,transGenoCMC_c,geneName,fullEngine,method);
	       }
	    }

	    if(WALD)
	    {
	       if(MB)
	       {
		  method = "MB";
		  WaldAssoc(fullTrans,MBWALDOut,transGenoMB,geneName,fullEngine,method);
	       }
	       if(CMC_binary)
	       {
		  method = "CMC_binary";	
		  WaldAssoc(fullTrans,CMC_bWALDOut,transGenoCMC_b,geneName,fullEngine,method);
	       }
	       if(CMC_counts)
	       {
		  method = "CMC_counts";	
		  WaldAssoc(fullTrans,CMC_cWALDOut,transGenoCMC_c,geneName,fullEngine,method);
	       }
	    }
	 }

	 String method;
	 if(LinScoreAssoc)
	 {
	    if(MB)
	    {
	       method = "MB";
	       LinAssoc(trans,engine,MBLinScoreOut,genotypeMB,geneName,method);
	    }
	    if(CMC_binary)
	    {
	       method = "CMC_binary";
	       LinAssoc(trans,engine,CMC_bLinScoreOut,genotypeCMC_b,geneName,method);
	    }
	    if(CMC_counts)
	    {
	       method = "CMC_counts";
	       LinAssoc(trans,engine,CMC_cLinScoreOut,genotypeCMC_c,geneName,method);
	    }
	 }

	 if(SCORE)
	 {
	    /*there are Merlin type score test */
	    if(MB)
	    {
	       method = "MB";
	       MerlinFastAssoc(trans,MBScoreOut,transGenoMB,geneName,engine,method);
	    }
	    if(CMC_binary)
	    {
	       method = "CMC_binary";
	       MerlinFastAssoc(trans,CMC_bScoreOut,transGenoCMC_b,geneName,engine,method);
	    }
	    if(CMC_counts)
	    {
	       method = "CMC_counts";
	       MerlinFastAssoc(trans,CMC_cScoreOut,transGenoCMC_c,geneName,engine,method);
	    }
	 }
	 if(SKAT)
	 {
	    SKATScoreAssoc(trans,SKATOut,geneName,engine,ped);
	 }
      } 
      else 
      {
	 bad++;
	 //printf("bad gene is:%s\n",group.annoGroups[i].c_str());
	 if(SCORE)
	 {
	    if(MB)
	    {
	       if(fullResult)
		  fprintf(MBScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(MBScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_binary)
	    {
	       if(fullResult)
		  fprintf(CMC_bScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(CMC_bScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_counts)
	    {
	       if(fullResult)
		  fprintf(CMC_cScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(CMC_cScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	 }
	 if(LinScoreAssoc)
	 {
	    if(MB)
	    {
	       if(fullResult)
		  fprintf(MBLinScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(MBLinScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_binary)
	    {
	       if(fullResult)
		  fprintf(CMC_bLinScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(CMC_bLinScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_counts)
	    {
	       if(fullResult)
		  fprintf(CMC_cLinScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(CMC_cLinScoreOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	 }
	 if(LRT)
	 {
	    if(MB)
	    {
	       if(fullResult)
		  fprintf(MBLRTOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(MBLRTOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_binary)
	    {
	       if(fullResult)
		  fprintf(CMC_bLRTOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(CMC_bLRTOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_counts)
	    {
	       if(fullResult)
		  fprintf(CMC_cLRTOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(CMC_cLRTOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	 }
	 if(WALD)
	 {
	    if(MB)
	    {
	       if(fullResult)
		  fprintf(MBWALDOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(MBWALDOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_binary)
	    {
	       if(fullResult)
		  fprintf(CMC_bWALDOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(CMC_bWALDOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_counts)
	    {
	       if(fullResult)
		  fprintf(CMC_cWALDOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	       else
		  fprintf(CMC_cWALDOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	 }
	 if(SKAT)
	 {
	    if(fullResult)
	       fprintf(SKATOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    else
	       fprintf(SKATOut,"%s\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	 }
      }
   }
   if(SCORE)
   {
      if(MB)
	 PrintFooter(MBScoreOut,bad,group);
      if(CMC_binary)
	 PrintFooter(CMC_bScoreOut,bad,group);
      if(CMC_counts)
	 PrintFooter(CMC_cScoreOut,bad,group);
   }
   if(LinScoreAssoc)
   {
      if(MB)
	 PrintFooter(MBLinScoreOut,bad,group);
      if(CMC_binary)
	 PrintFooter(CMC_bLinScoreOut,bad,group);
      if(CMC_counts)
	 PrintFooter(CMC_cLinScoreOut,bad,group);
   }
   if(LRT)
   {
      if(MB)
	 PrintFooter(MBLRTOut,bad,group);
      if(CMC_binary)
	 PrintFooter(CMC_bLRTOut,bad,group);
      if(CMC_counts)
	 PrintFooter(CMC_cLRTOut,bad,group);
   }
   if(WALD)
   {
      if(MB)
	 PrintFooter(MBWALDOut,bad,group);
      if(CMC_binary)
	 PrintFooter(CMC_bWALDOut,bad,group);
      if(CMC_counts)
	 PrintFooter(CMC_cWALDOut,bad,group);
   }
   if(SKAT)
      PrintFooter(SKATOut,bad,group);

   if(MB)
   {
      if(SCORE) fclose(MBScoreOut);
      if(LinScoreAssoc) fclose(MBLinScoreOut);
      if(LRT) fclose(MBLRTOut);
      if(WALD) fclose(MBWALDOut);
   }
   if(CMC_binary)
   {
      if(SCORE) fclose(CMC_bScoreOut);
      if(LinScoreAssoc) fclose(CMC_bLinScoreOut);
      if(LRT) fclose(CMC_bLRTOut);
      if(WALD) fclose(CMC_bWALDOut);
   } 
   if(CMC_counts)
   {
      if(SCORE) fclose(CMC_cScoreOut);
      if(LinScoreAssoc) fclose(CMC_cLinScoreOut);
      if(LRT) fclose(CMC_cLRTOut);
      if(WALD) fclose(CMC_cWALDOut);
   }
   if(SKAT)
      fclose(SKATOut);
   //Restoring the pedigree covariate settings.
   if((AutoFit::fitX || AutoFit::fitSharedEnvironment) && (LRT || WALD))
   {
      ped.covariateNames.Delete(ped.covariateNames.Length()-1);
      ped.covariateLookup.Delete("MarkerTested");
      ped.covariateCount--;
      for(int i=0;i<ped.count;i++)
	 ped[i].covariates[ped.covariateCount] = 0.0;
   }
}

void Burden::Run(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group,FILE * log)
{
if(SKAT || CMC_binary|| CMC_counts||MB)
burden= true;
   if(burden || SKAT)
   {
      if(!vcfAnnotated)
      {
	 error("\nERROR! VCF file needs to be annotated in INFO field.\n");
      }

      fixedEff.Dimension(trans.persons);
      //	 beta_hat.Dimension(trans.fixedEffects);
      //beta_hat.Copy(engine.betaHat);
      //engine.EstimateAnalyticalBeta(trans.UDY,trans, beta_hat);
      for(int i=0;i<trans.persons;i++)
      {
	 fixedEff[i] = trans.UDX[i].InnerProduct(engine.betaHat);
      }
      residuals_original.Copy(trans.UDY);
      residuals_original.Subtract(fixedEff);

      //sigma2_hat = residuals_original.SumSquares()/trans.persons;
      //printf("sigma2_hat is:%g\n",sigma2_hat);
      sigma2_hat =1.0;

      if(SKAT)
      {
	 printf("Start doing SKAT ...\n");
	 P0.Dimension(trans.persons,trans.persons);
	 Matrix tmp;
	 tmp.Product(trans.UDX,trans.inv);
	 Matrix X;
	 X.Transpose(trans.UDX);
	 P0.Product(tmp,X);
	 for(int i=0;i<trans.persons;i++)
	 {
	    P0[i][i] = 1.0 - P0[i][i];
	    //P0[i][i] *= sigma2_hat;
	    for(int j=i+1;j<trans.persons;j++)
	    {
	       P0[i][j] = -P0[i][j]; 
	       P0[j][i] = P0[i][j]; 
	    }
	 }
	 /*
	    printf("P0 is\n");
	    for(int i=0;i<trans.persons;i++)
	    {
	    for(int j=0;j< trans.persons;j++)
	    {
	    printf("%g ",P0[i][j]);
	    }
	    printf("\n");
	    }
	  */
	 printf("SKAT initialization done. \n");
      }

      if(MB ||CMC_binary||CMC_counts)
	 printf("Start doing the following Burden test: ");

      if(MB)
      {
	 printf("Madsen-Browning ");
      }

      if(CMC_binary)
      {
	 printf("CMC_binary ");
      }

      if(CMC_counts)
      {
	 printf("CMC_counts ");
      }
      printf(".\n");

      if(SKAT)
	 printf("Start doing the SKAT score test ...\n");

      Calculate(ped,trans,engine,traitName,group,log);
   }
}

void Burden::LinAssoc(FastTransform & trans,FastFit & engine,FILE * out, Vector & genotype,String & geneName,String & method)
{
   //Fill in UDX, UD, and UDY
   //trans.FinalizeTransform(engine.sigma_g2Hat, engine.sigma_e2Hat);
   Vector transGeno;
   transGeno.Product(trans.UD,genotype);
   double pvalue,numerator,denominator,chisq,effSize;
   Vector denominator_2;
   denominator_2.Dimension(trans.fixedEffects);
   numerator = 0.0;
   denominator = 1.0e-10;
   denominator_2.Zero();

   for(int i=0;i<trans.persons;i++)
   {
      numerator += residuals_original[i]*transGeno[i];
      denominator += transGeno[i]*transGeno[i];
      for(int j=0;j<trans.fixedEffects;j++)
      {
	 denominator_2[j] += transGeno[i]*trans.UDX[i][j];
      }
   }

   Vector tmp; //tmp is T(sum_SiZi)*inv
   tmp.Dimension(trans.fixedEffects);

   for(int i=0;i<trans.fixedEffects;i++)
   {
      tmp[i] = trans.inv[i].InnerProduct(denominator_2);
   }

   denominator -= tmp.InnerProduct(denominator_2);
   denominator *= sigma2_hat;

   /*
      Vector U;
      U.Dimension(trans.persons);
   //fill in U for variance
   for(int p=0;p<trans.persons;p++)
   {
   U[p] = residuals_original[p] *(transGeno[p] - tmp.InnerProduct(trans.UDX[p]));
   }
   double var = U.SumSquares();
    */
   if(denominator==0) chisq=0.0;
   else chisq = numerator*numerator/denominator;
   pvalue = pchisq(chisq, 1,0,0);
   //pvalue = chidist(chisq, 1);
   effSize = numerator/denominator;
   if(method == "MB")
   {
      String eff;
      eff = "/sqrt(pq)";
      //effSize /= (SNPCount*1/)
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.Length();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.Length();m++)
	    fprintf(out,"%g,",singlevarEff[m]);

	 fprintf(out,"\t%s\t%g\n",eff.c_str(),pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,eff.c_str(),pvalue);
   }
   else
   {
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.Length();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.Length();m++)
	    fprintf(out,"%g,",singlevarEff[m]);

	 fprintf(out,"\t%g\t%g\n",effSize,pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,effSize,pvalue);
   }
}

void Burden::MerlinLRT(double L0,FastTransform & fullTrans,FILE * out, Vector & transGeno,String & geneName,FastFit & fullEngine,String & method)
{
   double L1,chisq,effSize,pvalue;
   if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
   {
      L1=fullEngine.EvaluateMulti(fullTrans,false);

      chisq = 2.0*(L1-L0);
      if(chisq<0)
	 chisq = -chisq;
      //printf("L0, L1, chisq = %g, %g, %g\n",L0,L1,chisq);
      pvalue = pchisq(chisq, 1,0,0);
      //pvalue = chidist(chisq, 1);
      effSize = fullEngine.betaHat[fullTrans.fixedEffects-1];
   }
   else 
   {
      for(int row=0;row<fullTrans.persons;row++)
      {
	 fullTrans.UX[row][fullTrans.fixedEffects-1] =transGeno[row];
      }
      fullEngine.OptimizeDelta(0.0001,fullTrans);
      //fullEngine.deltaHat = fullEngine.min;
      fullEngine.GetBeta(fullEngine.deltaHat,fullTrans);
      L1 = fullEngine.Evaluate(fullEngine.deltaHat,fullTrans,false);

      chisq = 2.0*(L1-L0);
      //printf("L0 is:%g,L1 is:%g,chisq is: %g\n",L0,L1,chisq);
      if(chisq<0)
	 chisq = -chisq;
      //printf("L0 is:%g,L1 is:%g,chisq is: %g\n",L0,L1,chisq);
      pvalue = pchisq(chisq,1,0,0);
      //pvalue = chidist(chisq,1);
      effSize = fullEngine.beta[fullTrans.fixedEffects-1];
   }
   if(method == "MB")
   {
      String eff;
      eff = "/sqrt(pq)";
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.Length();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.Length();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%s\t%g\n",eff.c_str(),pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,eff.c_str(),pvalue);
   }
   else
   {
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.Length();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.Length();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%g\t%g\n",effSize,pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,effSize,pvalue);
   }
}

//Genotype here should be the transformed genotype
void Burden::WaldAssoc(FastTransform & fullTrans,FILE * out, Vector & genotype,String & geneName,FastFit & fullEngine,String & method)
{
   double pvalue,chisq,effSize;
   if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
   {
      fullTrans.UDX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
      fullTrans.UDY.Dimension(fullTrans.persons);
      fullTrans.inv.Dimension(fullTrans.fixedEffects,fullTrans.fixedEffects);

      fullTrans.FinalizeTransform(0.0,0.0);
      effSize = fullEngine.betaHat[fullTrans.fixedEffects-1];
      chisq = effSize*effSize/fullTrans.inv[fullTrans.fixedEffects-1][fullTrans.fixedEffects-1];
      pvalue = pchisq(chisq,1,0,0);
      //pvalue = chidist(chisq,1);
   }
   else
   {
      for(int row=0;row<fullTrans.persons;row++)
      {
	 fullTrans.UX[row][fullTrans.fixedEffects-1] = genotype[row];
      }
      fullEngine.OptimizeDelta(0.0001,fullTrans);
      //fullEngine.deltaHat = fullEngine.min;
      fullEngine.GetBeta(fullEngine.deltaHat,fullTrans);
      effSize = fullEngine.beta[fullTrans.fixedEffects-1];
      fullEngine.sigma_g2Hat = fullEngine.GetSigma_g(fullEngine.deltaHat,fullEngine.beta,fullTrans);
      fullEngine.sigma_e2Hat = fullEngine.sigma_g2Hat * fullEngine.deltaHat;

      fullTrans.UDX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
      fullTrans.UDY.Dimension(fullTrans.persons);
      fullTrans.inv.Dimension(fullTrans.fixedEffects,fullTrans.fixedEffects);

      fullTrans.FinalizeTransform(fullEngine.sigma_g2Hat,fullEngine.sigma_e2Hat);
      chisq = effSize*effSize/fullTrans.inv[fullTrans.fixedEffects-1][fullTrans.fixedEffects-1];
      pvalue = pchisq(chisq,1,0,0); 
      //pvalue = chidist(chisq,1); 
   }
   //double pvalue = 2.0*tdist(t0, fullTrans.persons-fullTrans.fixedEffects);
   if(method == "MB")
   {
      String eff; 
      eff = "/sqrt(pq)";
      if(fullResult)
      {    
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.Length();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.Length();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%s\t%g\n",eff.c_str(),pvalue);
      }    
      else 
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,eff.c_str(),pvalue);
   }
   else
   {
      if(fullResult)
      {    
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.Length();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.Length();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%g\t%g\n",effSize,pvalue);
      }    
      else 
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,effSize,pvalue);
   }
}

void Burden::CalculateSingleVarPvalue(FastTransform & trans,String & geneName,FastFit & engine,Matrix & genotypeSKAT,Pedigree & ped)
{
   //printf("Working on singlevar on gene %s.\n",geneName.c_str());
   Matrix transGeno;
   transGeno.Dimension(genotypeSKAT.rows,genotypeSKAT.cols);
   if(!trans.empKin)
   {
      transGeno.Zero();
      for(int m=0;m<genotypeSKAT.cols;m++) 
      {
	 int index=0;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    if (trans.pPheno[f].Length() > 0)
	    {
	       int count = trans.pPheno[f].Length();
	       int block = index;
	       for(int i=0;i<count;i++)
	       {
		  for(int c=0;c<count;c++)
		  {
		     transGeno[index][m] += trans.UD[block+i][block+c]*genotypeSKAT[block+c][m];
		  }
		  index++;
	       }
	    }
	 }
      }
   } else transGeno.Product(trans.UD,genotypeSKAT);

   // Calculate singlevariant pvalues
   singlevarPvalue.Dimension(genotypeSKAT.cols,0.0);
   singlevarEff.Dimension(genotypeSKAT.cols,0.0);
   for(int m=0;m<genotypeSKAT.cols;m++)
   {
      Vector geno;
      geno.Dimension(trans.persons);
      for(int p =0;p<trans.persons;p++)
      {
	 geno[p] = transGeno[p][m];
      }
      double numerator = residuals_original.InnerProduct(geno);
      double denominator = geno.InnerProduct(geno);
      double pvalue =  pchisq(numerator*numerator/denominator,1,0,0);
      //double pvalue =  chidist(numerator*numerator/denominator,1);
      singlevarPvalue[m] = pvalue;
      singlevarEff[m] = numerator/denominator;
   }
   //printf("finished working on singlevar on gene %s.\n",geneName.c_str());
}

void Burden::SKATScoreAssoc(FastTransform & trans,FILE * out,String & geneName,FastFit & engine,Pedigree & ped)
{
   //delete the columms that are redundant
   double davies_pvalue,liu_pvalue;
   bool disect=false;
   mafSKATIncluded.Copy(mafIncluded);
   if(del_status.Length()>0)
   {
      //printf("del_status.Length is:%d\n",del_status.Length());
      for(int i=del_status.Length()-1;i>=0;i--)
      {
	 if(del_status[i]==1)
	 {
	    genotypeSKAT.DeleteColumn(i);
	    mafSKATIncluded.Delete(i);
	 }
      }
   }
   //printf("there are %d variants included in SKAT\n",genotypeSKAT.cols);
   //printf("Doing SKAT on gene %s.\n",geneName.c_str());
   if(genotypeSKAT.cols==1)
   {
      Vector transGeno;
      transGeno.Dimension(trans.persons,0.0);
      for(int i=0;i<trans.persons;i++)
      {
	 for(int j=0;j<trans.persons;j++)
	    transGeno[i] += trans.UD[i][j]*genotypeSKAT[j][0];
      }
      double numerator = residuals_original.InnerProduct(transGeno);
      double denominator = transGeno.InnerProduct(transGeno);
      double chisq = numerator*numerator/denominator;
      davies_pvalue = pchisq(chisq,1,0,0);
      //davies_pvalue = chidist(chisq,1);
      liu_pvalue = davies_pvalue;
   }
   else
   {
      Matrix transGeno;
      transGeno.Dimension(genotypeSKAT.rows,genotypeSKAT.cols);
      if(!trans.empKin)
      {
	 transGeno.Zero();
	 for(int m=0;m<genotypeSKAT.cols;m++)
	 {
	    int index=0;
	    for (int f = 0; f < ped.familyCount; f++)
	    {
	       if (trans.pPheno[f].Length() > 0)
	       {
		  int count = trans.pPheno[f].Length();
		  int block = index;
		  for(int i=0;i<count;i++)
		  {
		     for(int c=0;c<count;c++)
		     {
			transGeno[index][m] += trans.UD[block+i][block+c]*genotypeSKAT[block+c][m];
		     }
		     index++;
		  }
	       }
	    }
	 }
      } else transGeno.Product(trans.UD,genotypeSKAT);

      double Qstat,pvalue;

      Matrix GWG;
      GWG.Dimension(trans.persons,trans.persons);
      GWG.Zero();
      Vector weight;
      weight.Dimension(genotypeSKAT.cols);

      for(int i=0;i<weight.Length();i++)
      {
	 double alpha = 1.0;
	 double beta=25.0;
	 double beta_density = GetBetaDensity(alpha,beta,mafSKATIncluded[i]);
	 weight[i] = beta_density * beta_density;
      }

      for(int i=0;i<trans.persons;i++)
      {
	 for(int m=0;m<genotypeSKAT.cols;m++)
	 {
	    GWG[i][i] += weight[m]*transGeno[i][m]*transGeno[i][m];
	 }
	 for(int j=i+1;j<trans.persons;j++)
	 {
	    for(int m=0;m<genotypeSKAT.cols;m++)
	    {
	       GWG[i][j] += transGeno[i][m]*transGeno[j][m]*weight[m];
	    }
	    GWG[j][i] = GWG[i][j];
	 }
      }

      Vector tmp;
      tmp.Dimension(trans.persons);
      for(int i=0;i<tmp.Length();i++)
      {
	 tmp[i] = residuals_original.InnerProduct(GWG[i]);
      }
      Qstat = tmp.InnerProduct(residuals_original);
      /*
	 printf("Qstat is %g; number of variant is %d\n",Qstat,weight.Length());
	 printf("weights are:\n");
	 for(int i=0;i<weight.Length();i++)
	 printf("%g ",weight[i]);
	 printf("\n");
       */
      int n = weight.Length();
      double * lambda = new double [n];

      CalculateLambda(weight,lambda,transGeno); //now lambda is filled in
      /*
	 printf("Lambdas are:\n");
	 for(int i=0;i<weight.Length();i++)
	 printf("%g ",lambda[i]);
	 printf("\n");
       */
      pvalue = MixChidist(lambda, n, Qstat,"Davies");
      /*
	 if(pvalue<=0 ||pvalue==2.0 || isnan(pvalue))
	 {
	 pvalue = MixChidist(lambda, n, Qstat,"Liu");
	 }
       */
      liu_pvalue = MixChidist(lambda, n, Qstat,"Liu");
      int disect_itr=0;
      while( (pvalue<=0 ||pvalue==2.0 || isnan(pvalue)) && disect_itr<1000)
      {
	 disect=true;
	 Qstat*=0.9999;
	 pvalue = MixChidist(lambda, n, Qstat,"Davies");
	 disect_itr++;
      }
      disect_itr=0;
      while((pvalue<=0 ||pvalue==2.0 || isnan(pvalue))  && disect_itr<1000)
      {
	 Qstat*=0.99;
	 pvalue = MixChidist(lambda, n, Qstat,"Davies");
	 disect_itr++;
      }
      davies_pvalue = pvalue;
      if(lambda) delete [] lambda;
   }
   if(disect)
   {
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.Length();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.Length();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t<%g\t%g\n",davies_pvalue,liu_pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t<%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,davies_pvalue,liu_pvalue);
   }
   else
   {
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.Length();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.Length();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%g\t%g\n",davies_pvalue,liu_pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,davies_pvalue,liu_pvalue);
   }
}

double Burden::GetBetaDensity(double a, double b, double x)
{
   double density;
   density = exp(gammln(a+b)-gammln(a)-gammln(b)+(a-1.0)*log(x)+(b-1.0)*log(1.0-x));
   return density;
}

void Burden::CalculateLambda(Vector & weight,double lambda [],Matrix & transGeno)
{
   //Goal of this function is to fill in lambda
   //now we only take diagnol weight matrix 
   Matrix WG;
   WG.Dimension(genotypeSKAT.cols,genotypeSKAT.rows);
   for(int i=0;i<WG.rows;i++)
   {
      for(int j=0;j<WG.cols;j++)
      {
	 WG[i][j] = transGeno[j][i] * sqrt(weight[i]);	
      }
   }
   Matrix GW;
   GW.Transpose(WG);
   Matrix tmp;
   tmp.Product(WG,P0);
   Matrix final;
   final.Product(tmp,GW);
/*
      printf("final matrix is: \n");
      for(int i=0;i<final.rows;i++)
      {
      for(int j=0;j<final.cols;j++)
      printf("%g ",final[i][j]);
      printf("\n");
      }
 SVD svd;
   svd.Decompose(final);
   //Fill in lambda here
   for(int i=0;i<weight.Length();i++)
   {
      lambda[i] = svd.w[i];
   }
*/

   //use EIGEN library
   Eigen::MatrixXf final_eigen(tmp.rows,GW.cols);
   for(int i=0;i<tmp.rows;i++)
      for(int j=0;j<GW.cols;j++)
	 final_eigen(i,j) = final[i][j];

Eigen::JacobiSVD<Eigen::MatrixXf> svd(final_eigen, Eigen::ComputeThinU);

   const Eigen::VectorXf S = svd.singularValues();

   for(int i=0;i<weight.Length();i++)
   {
      lambda[i] = S[i];
   }
}

void Burden::MerlinFastAssoc(FastTransform & trans,FILE * out, Vector & transGeno,String & geneName,FastFit & engine,String & method)
{
   //printf("Now working on %s in gene %s\n",method.c_str(),geneName.c_str());
   double numerator = 0.0;
   double denominator = 0.0;
   /*
      double fixedEffect, chisq,pvalue,effSize;
      Vector sigma2;
      if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
      {
      sigma2.Copy(trans.D);
      }
      else 
      {
      sigma2.Dimension(trans.persons);
      for(int i=0;i<trans.persons;i++)
      sigma2[i] = engine.sigma_g2Hat*trans.D[i] + engine.sigma_e2Hat;
      }
      printf("engine.g2,e2,betahat:%g %g %g %g %g %g\n",engine.sigma_g2Hat,engine.sigma_e2Hat,engine.betaHat[0],engine.betaHat[1],engine.betaHat[2],engine.betaHat[3]);
      printf("sigma2 is:\n");
      for(int i=0;i<sigma2.Length();i++)
      printf("%g ",sigma2[i]);
      printf("\nnumerator and denominator: \n");
      for(int i=0;i<trans.persons;i++)
      {
      fixedEffect=trans.UX[i].InnerProduct(engine.betaHat);
      numerator += (trans.UY[i]-fixedEffect)*transGeno[i]/sigma2[i];
      denominator += transGeno[i]*transGeno[i]/sigma2[i];
   //    printf("%d %g %g %g %g %g\n",i,fixedEffect,genotype[i],transGeno[i],numerator,denominator);
   }
    */

   numerator = residuals_original.InnerProduct(transGeno);
   denominator = transGeno.InnerProduct(transGeno);
   double chisq,effSize,pvalue;
   //printf("numerator and denominator are: %g, %g\n",numerator,denominator);

   if(denominator==0) chisq=0.0;
   else chisq = numerator*numerator/denominator;
   pvalue = pchisq(chisq,1,0,0);
   //pvalue = chidist(chisq,1);
   effSize = numerator/denominator;
   //Adjust effSize to make sense
   if(method == "MB")
   {
      String eff;
      eff = "/sqrt(pq)";
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.Length();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.Length();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%g%s\t%g\n",effSize,eff.c_str(),pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g%s\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,effSize,eff.c_str(),pvalue);
   }
   else
   {
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.Length();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.Length();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%g\t%g\n",effSize,pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,effSize,pvalue);
   }
}

void Burden::PrintFooter(FILE * output,int bad,GroupFromAnnotation & group)
{
   fprintf(output,"===================================================================\n");
   fprintf(output,"\n%d %s tested.\n",group.geneCount,group.geneCount>1?"genes were":"gene was");
   fprintf(output,"There %s %d %s either had no rare allele in the qualified sample or did not pass the collapsing criterion.\n",bad>1?"are":"is", bad,bad>1?"genes":"gene");
}

void Burden::PrintHeader(FILE * output,FastTransform & trans,FastFit & engine,String & traitName)
{
   double h2 = 100.0*engine.sigma_g2Hat/trans.traitVar;
   double h2X = 100.0*engine.sigma_gXHat/trans.traitVar;
   double sharedEnv = 100.0* engine.sigma_s2Hat/trans.traitVar;

   if(!AutoFit::fitSharedEnvironment && !AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2);
   }
   else if(AutoFit::fitSharedEnvironment && AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, Xheritability = %6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X,sharedEnv);
   }
   else if(AutoFit::fitSharedEnvironment)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,sharedEnv);
   }
   else if(AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, Xheritability = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X);
   }

   fprintf(output,"===================================================================\n");

   if(fullResult)
      fprintf(output,"Gene\tNumVarInGene\tInformative_AC\tAvgBurden\tFreqMin\tFreqMax\tFreqMean\tFreqSE\tVarsIncluded\tMafIncluded\tSingleVarPvalue\tSingleVarEffSize\tEffSIZE\tpvalue\n");
   else
      fprintf(output,"Gene\tNumVarInGene\tInformative_AC\tAvgBurden\tFreqMin\tFreqMax\tFreqMean\tFreqSE\tEffSIZE\tpvalue\n");
}
void Burden::PrintSKATHeader(FILE * output,FastTransform & trans,FastFit & engine,String & traitName)
{
   double h2 = 100.0*engine.sigma_g2Hat/trans.traitVar;
   double h2X = 100.0*engine.sigma_gXHat/trans.traitVar;
   double sharedEnv = 100.0* engine.sigma_s2Hat/trans.traitVar;

   if(!AutoFit::fitSharedEnvironment && !AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2);
   }
   else if(AutoFit::fitSharedEnvironment && AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, Xheritability = %6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X,sharedEnv);
   }
   else if(AutoFit::fitSharedEnvironment)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,sharedEnv);
   }
   else if(AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, Xheritability = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X);
   }

   fprintf(output,"===================================================================\n");
   if(fullResult)
      fprintf(output,"Gene\tNumVarInGene\tInformative_AC\tAvgBurden\tFreqMin\tFreqMax\tFreqMean\tFreqSE\tVarsIncluded\tMafsIncluded\tSingleVarPvalue\tSingleVarEffSize\tDaviesPvalue\tLiuPvalue\n");
   else
      fprintf(output,"Gene\tNumVarInGene\tInformative_AC\tAvgBurden\tFreqMin\tFreqMax\tFreqMean\tFreqSE\tDaviesPvalue\tLiuPvalue\n");
}

