#include "Parameters.h"
#include "Pedigree.h"
#include "math.h"
#include <iostream>
#include "FastFit.h"
#include "TraitTransformations.h"
#include "OutputKin.h"
#include "TransformResiduals.h"
#include "PEDtoVCF.h"
#include "VT.h"
#include "Burden.h"
#include "SingleVariantTest.h"
#include <stdio.h>
#include <time.h>
#include "Error.h"
#include "AutoFit.h"
#include "GroupFromAnnotation.h"
#define MATHLIB_STANDALONE
#include "Rmath.h"
#include "DataQC.h"

int main(int argc, char ** argv)
{
   String pedfile, datfile, mapfile;
   double tol = 0.0001;

   BEGIN_LONG_PARAMETERS(additional)
      LONG_PARAMETER_GROUP("Kinship Options")
      LONG_PARAMETER("kinGeno", &FastTransform::empKin)
      LONG_DOUBLEPARAMETER("empMaf", &KinshipEmp::q)
      LONG_DOUBLEPARAMETER("empMiss", &KinshipEmp::miss)
      LONG_PARAMETER("outputX", &OutputKin::outputKinX)
      LONG_PARAMETER("outputKin", &OutputKin::outputKin)
      LONG_STRINGPARAMETER("kinFile", &FastTransform::readInEmp)
      LONG_STRINGPARAMETER("kinPrefix", &OutputKin::filePrefix)
      //LONG_PARAMETER_GROUP("Pre-processing Options")
      //LONG_PARAMETER("preProcess", &VT::preProcess)
      //LONG_INTPARAMETER("chopSize", &VT::chop)
      //LONG_STRINGPARAMETER("saveChopTo", &VT::chopTo)
      LONG_PARAMETER_GROUP("Input/Output Options")
      LONG_STRINGPARAMETER("vcf", &GroupFromAnnotation::vcfInput)
      LONG_STRINGPARAMETER("groupFile", &GroupFromAnnotation::groupFile)
      LONG_STRINGPARAMETER("freqFile", &GroupFromAnnotation::freqFile)
      LONG_STRINGPARAMETER("prefix", &Burden::outputFile)
      LONG_PARAMETER_GROUP("VC Options")
      LONG_PARAMETER("inverseNormal", &FastFit::inverseNormal)
      LONG_PARAMETER("fitSharedEnv", &AutoFit::fitSharedEnvironment)
      LONG_PARAMETER("fitX", &AutoFit::fitX)
      LONG_PARAMETER("useCovariates", &FastFit::useCovariates)
      LONG_STRINGPARAMETER("traitName", &FastFit::traitName)
      LONG_PARAMETER_GROUP("SingleVar Tests")
      LONG_PARAMETER("SingleVarLRT", &SVT::LRT)
      LONG_PARAMETER("SingleVarScore", &SVT::SCORE)
      LONG_PARAMETER("SingleVarWald", &SVT::WALD)
      LONG_PARAMETER_GROUP("Burden Tests")
      //LONG_PARAMETER("Burden", &Burden::burden)
      LONG_PARAMETER("SKAT", &Burden::SKAT)
      //LONG_PARAMETER("LinScore", &Burden::LinScoreAssoc)
      //LONG_PARAMETER("LRT", &Burden::LRT)
      //LONG_PARAMETER("Wald", &Burden::WALD)
      LONG_PARAMETER("MB", &Burden::MB)
      LONG_PARAMETER("CMC_binary", &Burden::CMC_binary)
      LONG_PARAMETER("CMC_counts", &Burden::CMC_counts)
      LONG_PARAMETER_GROUP("Variable Threshold Tests")
      LONG_PARAMETER("VTasymptotic", &VT::asymptotic)
      LONG_PARAMETER("VTpermute", &VT::permuteBool)
      LONG_INTPARAMETER("permuteMin", &VT::permuteMin)
      LONG_INTPARAMETER("permuteMax", &VT::permuteMax)
      //LONG_PARAMETER("VTMB", &VT::VTwtMB)
      //LONG_PARAMETER("VTCMC_binary", &VT::VTwtCMC_binary)
      //LONG_PARAMETER("VTCMC_counts", &VT::VTwtCMC_counts)
      LONG_PARAMETER_GROUP("Other Options")
      //LONG_STRINGPARAMETER("geneName", &Burden::genename)
      //LONG_STRINGPARAMETER("geneFile", &Burden::genename_f)
      LONG_STRINGPARAMETER("function", &GroupFromAnnotation::function)
      //LONG_STRINGPARAMETER("chrom", &Burden::chromosome)
      LONG_DOUBLEPARAMETER("mafMin", &Burden::mafMin)
      LONG_DOUBLEPARAMETER("mafMax", &Burden::mafMax)
      LONG_DOUBLEPARAMETER("mac", &Burden::MAC)
      LONG_PARAMETER("noStop", &VT::noStep)
      //LONG_INTPARAMETER("maleLabel", &SVT::maleLabel)
      //LONG_INTPARAMETER("femaleLabel", &SVT::femaleLabel)
      LONG_STRINGPARAMETER("xLabel", &SVT::xLabel)
      LONG_INTPARAMETER("Xstart", &SVT::Xstart)
      LONG_INTPARAMETER("Xend", &SVT::Xend)
      LONG_PARAMETER("dosage", &SVT::dosage)
      LONG_PARAMETER("founderFreq", &SVT::FounderFreq)
      LONG_PARAMETER("h2Only", &FastFit::h2Only)
      LONG_PARAMETER("fullResult", &Burden::fullResult)

      END_LONG_PARAMETERS();

   ParameterList pl;
   pl.Add(new StringParameter('d', "Data File", datfile));
   pl.Add(new StringParameter('p', "Pedigree File", pedfile));
   //    pl.Add(new StringParameter('m', "Map File", mapfile));
   pl.Add(new LongParameters("Options:", additional));

   pl.Read(argc, argv);

   pl.Status();

   /*
      if(!SVT::test && !Burden::MB && !Burden::CMC_counts && !Burden::CMC_binary)
      {
      error("ERROR! Rare variant test names need to be entered.\nOptions are: --svTest, --CMC_binary, --CMC_counts, and --MB.\n\n");
      }
    */
   Pedigree ped;
   ped.Prepare(datfile);
   ped.Load(pedfile);

   if(ped.markerCount>0)
      SVT::genoFromPed = true;
   else if(!SVT::dosage) 
      SVT::genoFromVCF=true;

   if(Burden::burden || Burden::SKAT || VT::asymptotic || VT::permuteBool)
   {
      if(GroupFromAnnotation::vcfInput =="" && GroupFromAnnotation::groupFile == "")
	 error("FATAL ERROR: annotated VCF file or group file needed for Burden tests.\n");
   }
   FILE * log;
String logFile;
   if(Burden::outputFile != "")
   {
      if(Burden::outputFile.FastFindLastChar('.')!=-1 || Burden::outputFile.FastFindLastChar('/')!=-1)
         logFile = Burden::outputFile + "famRvTest.log";
      else
         logFile = Burden::outputFile + ".famRvTest.log";
   }
   else
      logFile = "famRvTest.log";

log = freopen (logFile,"wt",stderr);


   SanityCheck checkData;
   checkData.Check(ped,log);

   AutoFit fitEngine;
   FastFit engine;
   FastTransform trans;
   KinshipEmp kin_emp;
   //check if the sample IDs from PED and VCF files are consistent
   trans.ScreenSampleID(ped,FastFit::useCovariates);

   //setup empirical kinship if applicable
   if(FastTransform::empKin)
      kin_emp.SetupEmpKin(ped,trans.genotypedSamplePED,trans.genotypedSampleVCF,trans.samplePEDIDHash,checkData.skippedSNPs,log);

   StringArray traits;
   if(!FastFit::traitName.IsEmpty())
   {
      if(FastFit::traitName == "traitName.txt")
      {
	 FILE * trFile = fopen(FastFit::traitName,"r");
	 String buffer;
	 while (!feof(trFile))
	 {
	    buffer.ReadLine(trFile);
	    traits.AddTokens(buffer, SEPARATORS);
	 }
	 fclose(trFile);
      }
      else
      {
	 traits.AddTokens(FastFit::traitName,"/");
      }
   }
   else 
   {
      traits = ped.traitNames;
   }

   for(int i=0;i<traits.Length();i++)
   {
      FastFit::traitName.Copy(traits[i]);
FILE * dummy;
      engine.FitModels(dummy,ped,tol,trans,kin_emp,log);

      if(FastFit::h2Only)
      {
	 FILE * h2out;
	 String file = Burden::outputFile + "_" + FastFit::traitName + "." + "heritability";
	 h2out=fopen(file,"wt");
	 double h2=engine.sigma_g2Hat/trans.traitVar;
	 printf("Heritability for trait %s saved in %s.\n",FastFit::traitName.c_str(),file.c_str());
	 fprintf(h2out,"%s\t%g\n",FastFit::traitName.c_str(),h2);
	 fclose(h2out);
	 continue;
      }

      //Now do the burden tests

      SVT svt;
      svt.Run(ped,trans,engine,traits[i],log);

      GroupFromAnnotation group;
      group.Run();

      if(Burden::SKAT || Burden::burden)
      {
	 Burden burden;
	 burden.Run(ped,trans,engine,traits[i],group,log);	
      }

      VT vt;
      vt.Run(ped,trans,engine,traits[i],group);

   }
fclose(log);
   return 0;
}
